/*
 * Created on Nov 22, 2004
 * Copyright (C) Andrea Schweer, 2004
 *
 * This file is part of the Greenstone Alerting Service.
 * Refer to the COPYING file in the base directory of this package
 * for licensing information.
 */
package org.greenstone.gsdlas;

import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

import javax.servlet.http.*;

import org.apache.velocity.Template;
import org.apache.velocity.context.Context;
import org.apache.velocity.servlet.VelocityServlet;
import org.greenstone.gsdlas.users.UserManagementException;
import org.greenstone.gsdlas.users.UserManager;

/**
 * @author schweer
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class AlertingService extends VelocityServlet {
    
    public static final Set actions;
    static {
        Set set = new TreeSet();
        set.add("createSubscription");
        set.add("deleteSubscription");
        set.add("editSubscription");
        set.add("showFeed");
        set.add("listSubscriptions");
        
        set.add("login");
        set.add("register");
        set.add("logout");
        set.add("showLoginForm");
        set.add("showRegistrationForm");
        actions = Collections.unmodifiableSet(set);
    }
    
    
    protected Template handleRequest(HttpServletRequest req,
            HttpServletResponse res, Context context) {
        
        String action = req.getParameter("action");
        
        if (action == null || !actions.contains(action)) {
            String title = "Unknown action";
            String message = "I don't know how to " + action;
            String details = "The only actions I know are " + actions;
            return showError(context, message, details);
        }
        

        Map args = req.getParameterMap();
        // TODO stop this, we need multi-valued stuff
        args = normalise(args);
        
        String templateString = "";
        
        try {
            Method method = AlertingService.class.getDeclaredMethod(action, new Class[] {Map.class, Context.class});
            templateString = (String) method.invoke(this, new Object[] {args, context});
        } catch (Exception e) {
            String message = "An error has occured, I couldn't do what you told me to do.";
            String details = e.getMessage() + " (" + e.getClass().getName() + "); "
            	+ e.getCause() 
            	+ " ; action is " + action;
            return showError(context, message, details);
        }
        
        Template template = null;
        try {
            template = getTemplate(templateString);
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            try {
                error((HttpServletRequest)context.get(REQUEST), 
                        (HttpServletResponse)context.get(RESPONSE),
                        e);
            } catch (Exception e2) {
                // TODO Auto-generated catch block
                e2.printStackTrace();
            }
        }
        return template;
    }
    
    /**
     * @param arguments
     * @param context
     * @return the Velocity template to use
     * @throws Exception
     */
    public String createSubscription(Map arguments, Context context) throws Exception {
        HttpSession session = ((HttpServletRequest)context.get(REQUEST)).getSession(true);
        if (!UserManager.getInstance().isLoggedIn(session)) {
            session.setAttribute("next_action", "createSubscription");
            return showLoginForm(arguments, context);
        }
        if (arguments.containsKey("next_page") && arguments.get("next_page").equals("finish")) {
            arguments.putAll(getPageArgsFromSession(session));
            ProfileStore.getInstance().createSubscription(arguments);
            return listSubscriptions(arguments, context);
        } else {
            return showSubscriptionWizardPage(arguments, context, true);
        }
    }

    public String deleteSubscription(Map arguments, Context context) {
        HttpSession session = ((HttpServletRequest)context.get(REQUEST)).getSession(true);
        if (!UserManager.getInstance().isLoggedIn(session)) {
            session.setAttribute("next_action", "deleteSubscription");
            return showLoginForm(arguments, context);
        }
        String subscriptionID = (String) arguments.get("subscriptionID");
        ProfileStore.getInstance().deleteSubscription(subscriptionID);
        return listSubscriptions(arguments, context);
    }
    
    public String editSubscription(Map arguments, Context context) throws Exception {
        HttpSession session = ((HttpServletRequest)context.get(REQUEST)).getSession(true);
        if (!UserManager.getInstance().isLoggedIn(session)) {
            session.setAttribute("next_action", "editSubscription");
            return showLoginForm(arguments, context);
        }
        if (arguments.containsKey("next_page") && arguments.get("next_page").equals("finish")) {
            ProfileStore.getInstance().changeSubscription(arguments, session);
            return listSubscriptions(arguments, context);
        } else {
            return showSubscriptionWizardPage(arguments, context, false);
        }
    }
    
    public String showFeed(Map arguments, Context context) {
        context.put("list", ""/* TODO pass subscription list */);
        return "feed.vm";
    }
    
    public String showLoginForm(Map arguments, Context context) {
        return "login.vm";
    }
    
    public String showRegistrationForm(Map arguments, Context context) {
        return "register.vm";
    }

    public String listSubscriptions(Map arguments, Context context) {
        HttpSession session = ((HttpServletRequest)context.get(REQUEST)).getSession(true);
        if (!UserManager.getInstance().isLoggedIn(session)) {
            session.setAttribute("next_action", "listSubscriptions");
            return showLoginForm(arguments, context);
        }
        String username = (String) session.getAttribute("username");
        context.put("title", "List of Subscriptions for " + username);
        context.put("list", ProfileStore.getInstance().getAllSubscriptions());
        return "list.vm";
    }
    
    public String login(Map arguments, Context context) throws Exception {
        HttpSession session = ((HttpServletRequest)context.get(REQUEST)).getSession(true);
        try {
            UserManager.getInstance().loginUser(arguments, session);
        } catch (UserManagementException e) {
            context.put("error", Boolean.TRUE);
            context.put("errormessage", e.getMessage());
            return showLoginForm(arguments, context); 
        }
        if (session.getAttribute("next_action") != null) {
            String nextAction = (String) session.getAttribute("next_action");
            Method method = AlertingService.class.getDeclaredMethod(nextAction, new Class[] {Map.class, Context.class});
            return (String) method.invoke(this, new Object[] {arguments, context});
        }
        return listSubscriptions(arguments, context);
    }
    
    public String register(Map arguments, Context context) throws Exception {
        HttpSession session = ((HttpServletRequest)context.get(REQUEST)).getSession(true);
        try {
            UserManager.getInstance().createUser(arguments, session);
        } catch (UserManagementException e) {
            context.put("error", Boolean.TRUE);
            context.put("errormessage", e.getMessage());
            return showRegistrationForm(arguments, context);
        }
        return login(arguments, context);
    }
    
    public String logout(Map arguments, Context context) {
        HttpSession session = ((HttpServletRequest)context.get(REQUEST)).getSession(true);
        Enumeration atts = session.getAttributeNames();
        while (atts.hasMoreElements()) {
            session.removeAttribute((String) atts.nextElement());
        }
        session.invalidate();
        return "general.vm";
    }
    
    public void receiveEvent(Map event) {
        GreenstoneCommunicator gsComm = null;
        try {
            gsComm = new GreenstoneCommunicator(new URL((String) event.get("host_url")));
        } catch (MalformedURLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        Set matchedSubscriptions = ProfileStore.getInstance().filter(event, gsComm);
        // TODO do something with the matched subscriptions
    }

    /**
     * @param args
     */
    private Map normalise(Map args) {
        Map result = new TreeMap();
        for (Iterator iter = args.keySet().iterator(); iter.hasNext();) {
            String key = (String) iter.next();
            if (key.equals("type") || key.equals("host") 
                    || key.equals("collection") || key.equals("way")) {
                // multi-valued attributes
                String[] values = ((String[]) args.get(key));
                result.put(key, Arrays.asList(values));
            } else {
                String firstValue = ((String[])args.get(key))[0];
                result.put(key, firstValue);
            }
        }
        return result;
    }

    /**
     * @param context
     * @param message
     * @param details
     * @return
     */
    private Template showError(Context context, String message, String details) {
        context.put("title", "Error");
        context.put("message", message);
        context.put("details", details);
        try {
            return getTemplate("error.vm");
        } catch (Exception e) {
            try {
                super.error((HttpServletRequest)context.get("req"), 
                        (HttpServletResponse)context.get("res"), 
                        e);
            } catch (Exception e2) {
                e2.printStackTrace();
            }
        }
        return null;
    }

    /**
     * @param arguments
     * @param context
     * @param create
     * @return
     * @throws Exception
     */
    private String showSubscriptionWizardPage(Map arguments, Context context, boolean create) throws Exception {
        if (create) {
            context.put("action", "createSubscription");
        } else {
            context.put("action", "editSubscription");
        }
        if (!arguments.containsKey("current_page")) {
            return "sub_type-details.vm";
        }
        
        HttpSession session = ((HttpServletRequest)context.get(REQUEST)).getSession(true);
        String currentPage = (String) arguments.get("current_page");
        String direction = (String) arguments.get("next_page");
        
        // save page arguments
        savePageArgsToSession(currentPage, arguments, session);
        
        String nextPage = getNextPage(currentPage, direction);

        // fill prefill
        context.put("prefill", getPageArgsFromSession(nextPage, session));
        
        // fill preview
        context.put("preview", getPagePreview(nextPage, session));
        
        // get page-specific stuff
        if (nextPage.equals("host")) {
            GreenstoneCommunicator gsComm = null;
            String[] hostNames;
            try {
                gsComm = new GreenstoneCommunicator();
                hostNames = gsComm.getHostNames();
            } catch (Exception e) {
                hostNames = new String[] { "localhost" };
            }
            context.put("hostnames", hostNames);
        } else if (nextPage.equals("collection")) {
            // TODO might be query instead of just name
            List hostNames = (List) arguments.get("host");
            GreenstoneCommunicator gsComm = null;
            Map collNames = new TreeMap();
            for (Iterator iter = hostNames.iterator(); iter.hasNext();) {
                String host = (String) iter.next();
                Set collNamesForHost = new TreeSet();   
                try {
                    gsComm = new GreenstoneCommunicator(new URL("http://" + host + ":8080/soap/servlet/rpcrouter"));
                    collNamesForHost.addAll(Arrays.asList(gsComm.getCollectionNames()));
                } catch (Exception e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }	
                collNames.put(host, collNamesForHost);
            }
            context.put("collectionnames", collNames);
            context.put("hostnames", hostNames);
        }
        
        return "sub_" + nextPage + ".vm";
    }

    /**
     * @param nextPage
     * @param session
     * @return
     */
    private Map getPagePreview(String nextPage, HttpSession session) {
        Map preview = new TreeMap();
        if (nextPage.equals("type-details")) {
            return preview;
        }
        preview.putAll(getPageArgsFromSession("type-details", session));
        if (nextPage.equals("host")) {
            return preview;
        }
        preview.putAll(getPageArgsFromSession("host", session));
        if (nextPage.equals("collection")) {
            return preview;
        }
        preview.putAll(getPageArgsFromSession("collection", session));
        return preview;
    }

    /**
     * @param currentPage
     * @param direction
     * @return
     * @throws Exception
     */
    private String getNextPage(String currentPage, String direction) throws Exception {
        String nextPage;
        if (currentPage.equals("host") && direction.equals("back")) {
            nextPage = "type-details";
        } else if (currentPage.equals("type-details") || (currentPage.equals("collection") && direction.equals("back"))) {
            nextPage = "host";
        } else if (currentPage.equals("host") || (currentPage.equals("notification") && direction.equals("back"))) {
            nextPage = "collection";
        } else if (currentPage.equals("collection")) {
            nextPage = "notification";
        } else {
            throw new Exception("unknown combination of currentPage=" + currentPage + " and nextPage=" + direction);
        }
        return nextPage;
    }

    /**
     * @param page
     * @param session
     * @return 
     */
    private Map getPageArgsFromSession(String page, HttpSession session) {
        Map pageArgs = (Map) session.getAttribute("page_args");
        if (pageArgs == null || !pageArgs.containsKey(page))
            return new TreeMap();
        return (Map) pageArgs.get(page);
    }

    private Map getPageArgsFromSession(HttpSession session) {
        Map result = new TreeMap();
        Map pageArgs = (Map) session.getAttribute("page_args");
        if (pageArgs != null) {
            for (Iterator iter = pageArgs.values().iterator(); iter.hasNext();) {
                Map args = (Map) iter.next();
                result.putAll(args);
            }
        }
        return result;
    }
    
    /**
     * @param page
     * @param arguments
     * @param session
     */
    private void savePageArgsToSession(String page, Map arguments, HttpSession session) {
        Map pageArgs = (Map) session.getAttribute("page_args");
        if (pageArgs == null) {
            pageArgs = new TreeMap();
        }
        pageArgs.put(page, arguments);
        session.setAttribute("page_args", pageArgs);
    }
}
